@@ -12,6 +12,7 @@
 #include <stack>
 #include <utility>
 #include <vector>
+#include <Eigen/Core>
 
 #include "alpha_division_effect.h"
 #include "alpha_multiplication_effect.h"
@@ -19,13 +20,16 @@
 #include "dither_effect.h"
 #include "effect.h"
 #include "effect_chain.h"
+#include "effect_util.h"
 #include "gamma_compression_effect.h"
 #include "gamma_expansion_effect.h"
 #include "init.h"
 #include "input.h"
 #include "resource_pool.h"
 #include "util.h"
+#include "ycbcr_conversion_effect.h"
 
+using namespace Eigen;
 using namespace std;
 
 namespace movit {
@@ -35,8 +39,10 @@ EffectChain::EffectChain(float aspect_nom, float aspect_denom, ResourcePool *res
 	  aspect_denom(aspect_denom),
 	  dither_effect(NULL),
 	  num_dither_bits(0),
+	  output_origin(OUTPUT_ORIGIN_BOTTOM_LEFT),
 	  finalized(false),
-	  resource_pool(resource_pool) {
+	  resource_pool(resource_pool),
+	  do_phase_timing(false) {
 	if (resource_pool == NULL) {
 		this->resource_pool = new ResourcePool();
 		owns_resource_pool = true;
@@ -73,6 +79,21 @@ void EffectChain::add_output(const ImageFormat &format, OutputAlphaFormat alpha_
 	assert(!finalized);
 	output_format = format;
 	output_alpha_format = alpha_format;
+	output_color_type = OUTPUT_COLOR_RGB;
+}
+
+void EffectChain::add_ycbcr_output(const ImageFormat &format, OutputAlphaFormat alpha_format,
+                                   const YCbCrFormat &ycbcr_format, YCbCrOutputSplitting output_splitting)
+{
+	assert(!finalized);
+	output_format = format;
+	output_alpha_format = alpha_format;
+	output_color_type = OUTPUT_COLOR_YCBCR;
+	output_ycbcr_format = ycbcr_format;
+	output_ycbcr_splitting = output_splitting;
+
+	assert(ycbcr_format.chroma_subsampling_x == 1);
+	assert(ycbcr_format.chroma_subsampling_y == 1);
 }
 
 Node *EffectChain::add_node(Effect *effect)
@@ -87,6 +108,8 @@ Node *EffectChain::add_node(Effect *effect)
 	node->output_color_space = COLORSPACE_INVALID;
 	node->output_gamma_curve = GAMMA_INVALID;
 	node->output_alpha_type = ALPHA_INVALID;
+	node->needs_mipmaps = false;
+	node->one_to_one_sampling = false;
 
 	nodes.push_back(node);
 	node_map[effect] = node;
@@ -226,11 +249,61 @@ string replace_prefix(const string &text, const string &prefix)
 	return output;
 }
 
+namespace {
+
+template<class T>
+void extract_uniform_declarations(const vector<Uniform<T> > &effect_uniforms,
+                                  const string &type_specifier,
+                                  const string &effect_id,
+                                  vector<Uniform<T> > *phase_uniforms,
+                                  string *glsl_string)
+{
+	for (unsigned i = 0; i < effect_uniforms.size(); ++i) {
+		phase_uniforms->push_back(effect_uniforms[i]);
+		phase_uniforms->back().prefix = effect_id;
+
+		*glsl_string += string("uniform ") + type_specifier + " " + effect_id
+			+ "_" + effect_uniforms[i].name + ";\n";
+	}
+}
+
+template<class T>
+void extract_uniform_array_declarations(const vector<Uniform<T> > &effect_uniforms,
+                                        const string &type_specifier,
+                                        const string &effect_id,
+                                        vector<Uniform<T> > *phase_uniforms,
+                                        string *glsl_string)
+{
+	for (unsigned i = 0; i < effect_uniforms.size(); ++i) {
+		phase_uniforms->push_back(effect_uniforms[i]);
+		phase_uniforms->back().prefix = effect_id;
+
+		char buf[256];
+		snprintf(buf, sizeof(buf), "uniform %s %s_%s[%d];\n",
+			type_specifier.c_str(), effect_id.c_str(),
+			effect_uniforms[i].name.c_str(),
+			int(effect_uniforms[i].num_values));
+		*glsl_string += buf;
+	}
+}
+
+template<class T>
+void collect_uniform_locations(GLuint glsl_program_num, vector<Uniform<T> > *phase_uniforms)
+{
+	for (unsigned i = 0; i < phase_uniforms->size(); ++i) {
+		Uniform<T> &uniform = (*phase_uniforms)[i];
+		uniform.location = get_uniform_location(glsl_program_num, uniform.prefix, uniform.name);
+	}
+}
+
+}  // namespace
+
 void EffectChain::compile_glsl_program(Phase *phase)
 {
-	string frag_shader = read_version_dependent_file("header", "frag");
+	string frag_shader_header = read_version_dependent_file("header", "frag");
+	string frag_shader = "";
 
-	// Create functions for all the texture inputs that we need.
+	// Create functions and uniforms for all the texture inputs that we need.
 	for (unsigned i = 0; i < phase->inputs.size(); ++i) {
 		Node *input = phase->inputs[i]->output_node;
 		char effect_id[256];
@@ -242,14 +315,27 @@ void EffectChain::compile_glsl_program(Phase *phase)
 		frag_shader += "\treturn tex2D(tex_" + string(effect_id) + ", tc);\n";
 		frag_shader += "}\n";
 		frag_shader += "\n";
+
+		Uniform<int> uniform;
+		uniform.name = effect_id;
+		uniform.value = &phase->input_samplers[i];
+		uniform.prefix = "tex";
+		uniform.num_values = 1;
+		uniform.location = -1;
+		phase->uniforms_sampler2d.push_back(uniform);
 	}
 
+	// Give each effect in the phase its own ID.
 	for (unsigned i = 0; i < phase->effects.size(); ++i) {
 		Node *node = phase->effects[i];
 		char effect_id[256];
 		sprintf(effect_id, "eff%u", i);
 		phase->effect_ids.insert(make_pair(node, effect_id));
+	}
 
+	for (unsigned i = 0; i < phase->effects.size(); ++i) {
+		Node *node = phase->effects[i];
+		const string effect_id = phase->effect_ids[node];
 		if (node->incoming_links.size() == 1) {
 			frag_shader += string("#define INPUT ") + phase->effect_ids[node->incoming_links[0]] + "\n";
 		} else {
@@ -262,7 +348,6 @@ void EffectChain::compile_glsl_program(Phase *phase)
 	
 		frag_shader += "\n";
 		frag_shader += string("#define FUNCNAME ") + effect_id + "\n";
-		frag_shader += replace_prefix(node->effect->output_convenience_uniforms(), effect_id);
 		frag_shader += replace_prefix(node->effect->output_fragment_shader(), effect_id);
 		frag_shader += "#undef PREFIX\n";
 		frag_shader += "#undef FUNCNAME\n";
@@ -278,17 +363,80 @@ void EffectChain::compile_glsl_program(Phase *phase)
 		frag_shader += "\n";
 	}
 	frag_shader += string("#define INPUT ") + phase->effect_ids[phase->effects.back()] + "\n";
+
+	// If we're the last phase, add the right #defines for Y'CbCr multi-output as needed.
+	if (phase->output_node->outgoing_links.empty() && output_color_type == OUTPUT_COLOR_YCBCR) {
+		switch (output_ycbcr_splitting) {
+		case YCBCR_OUTPUT_INTERLEAVED:
+			// No #defines set.
+			break;
+		case YCBCR_OUTPUT_SPLIT_Y_AND_CBCR:
+			frag_shader += "#define YCBCR_OUTPUT_SPLIT_Y_AND_CBCR 1\n";
+			break;
+		case YCBCR_OUTPUT_PLANAR:
+			frag_shader += "#define YCBCR_OUTPUT_PLANAR 1\n";
+			break;
+		default:
+			assert(false);
+		}
+	}
 	frag_shader.append(read_version_dependent_file("footer", "frag"));
 
+	// Collect uniforms from all effects and output them. Note that this needs
+	// to happen after output_fragment_shader(), even though the uniforms come
+	// before in the output source, since output_fragment_shader() is allowed
+	// to register new uniforms (e.g. arrays that are of unknown length until
+	// finalization time).
+	// TODO: Make a uniform block for platforms that support it.
+	string frag_shader_uniforms = "";
+	for (unsigned i = 0; i < phase->effects.size(); ++i) {
+		Node *node = phase->effects[i];
+		Effect *effect = node->effect;
+		const string effect_id = phase->effect_ids[node];
+		extract_uniform_declarations(effect->uniforms_sampler2d, "sampler2D", effect_id, &phase->uniforms_sampler2d, &frag_shader_uniforms);
+		extract_uniform_declarations(effect->uniforms_bool, "bool", effect_id, &phase->uniforms_bool, &frag_shader_uniforms);
+		extract_uniform_declarations(effect->uniforms_int, "int", effect_id, &phase->uniforms_int, &frag_shader_uniforms);
+		extract_uniform_declarations(effect->uniforms_float, "float", effect_id, &phase->uniforms_float, &frag_shader_uniforms);
+		extract_uniform_declarations(effect->uniforms_vec2, "vec2", effect_id, &phase->uniforms_vec2, &frag_shader_uniforms);
+		extract_uniform_declarations(effect->uniforms_vec3, "vec3", effect_id, &phase->uniforms_vec3, &frag_shader_uniforms);
+		extract_uniform_declarations(effect->uniforms_vec4, "vec4", effect_id, &phase->uniforms_vec4, &frag_shader_uniforms);
+		extract_uniform_array_declarations(effect->uniforms_vec2_array, "vec2", effect_id, &phase->uniforms_vec2, &frag_shader_uniforms);
+		extract_uniform_array_declarations(effect->uniforms_vec4_array, "vec4", effect_id, &phase->uniforms_vec4, &frag_shader_uniforms);
+		extract_uniform_declarations(effect->uniforms_mat3, "mat3", effect_id, &phase->uniforms_mat3, &frag_shader_uniforms);
+	}
+
+	frag_shader = frag_shader_header + frag_shader_uniforms + frag_shader;
+
 	string vert_shader = read_version_dependent_file("vs", "vert");
+
+	// If we're the last phase and need to flip the picture to compensate for
+	// the origin, tell the vertex shader so.
+	if (phase->output_node->outgoing_links.empty() && output_origin == OUTPUT_ORIGIN_TOP_LEFT) {
+		const string needle = "#define FLIP_ORIGIN 0";
+		size_t pos = vert_shader.find(needle);
+		assert(pos != string::npos);
+
+		vert_shader[pos + needle.size() - 1] = '1';
+	}
+
 	phase->glsl_program_num = resource_pool->compile_glsl_program(vert_shader, frag_shader);
+
+	// Collect the resulting location numbers for each uniform.
+	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_sampler2d);
+	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_bool);
+	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_int);
+	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_float);
+	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_vec2);
+	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_vec3);
+	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_vec4);
+	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_mat3);
 }
 
 // Construct GLSL programs, starting at the given effect and following
 // the chain from there. We end a program every time we come to an effect
 // marked as "needs texture bounce", one that is used by multiple other
-// effects, every time an effect wants to change the output size,
-// and of course at the end.
+// effects, every time we need to bounce due to output size change
+// (not all size changes require ending), and of course at the end.
 //
 // We follow a quite simple depth-first search from the output, although
 // without recursing explicitly within each phase.
@@ -301,6 +449,13 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 	Phase *phase = new Phase;
 	phase->output_node = output;
 
+	// If the output effect has one-to-one sampling, we try to trace this
+	// status down through the dependency chain. This is important in case
+	// we hit an effect that changes output size (and not sets a virtual
+	// output size); if we have one-to-one sampling, we don't have to break
+	// the phase.
+	output->one_to_one_sampling = output->effect->one_to_one_sampling();
+
 	// Effects that we have yet to calculate, but that we know should
 	// be in the current phase.
 	stack<Node *> effects_todo_this_phase;
@@ -310,6 +465,10 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 		Node *node = effects_todo_this_phase.top();
 		effects_todo_this_phase.pop();
 
+		if (node->effect->needs_mipmaps()) {
+			node->needs_mipmaps = true;
+		}
+
 		// This should currently only happen for effects that are inputs
 		// (either true inputs or phase outputs). We special-case inputs,
 		// and then deduplicate phase outputs below.
@@ -334,6 +493,21 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 				start_new_phase = true;
 			}
 
+			// Propagate information about needing mipmaps down the chain,
+			// breaking the phase if we notice an incompatibility.
+			//
+			// Note that we cannot do this propagation as a normal pass,
+			// because it needs information about where the phases end
+			// (we should not propagate the flag across phases).
+			if (node->needs_mipmaps) {
+				if (deps[i]->effect->num_inputs() == 0) {
+					Input *input = static_cast<Input *>(deps[i]->effect);
+					start_new_phase |= !input->can_supply_mipmaps();
+				} else {
+					deps[i]->needs_mipmaps = true;
+				}
+			}
+
 			if (deps[i]->outgoing_links.size() > 1) {
 				if (!deps[i]->effect->is_single_texture()) {
 					// More than one effect uses this as the input,
@@ -359,7 +533,14 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 				}
 			}
 
-			if (deps[i]->effect->changes_output_size()) {
+			if (deps[i]->effect->sets_virtual_output_size()) {
+				assert(deps[i]->effect->changes_output_size());
+				// If the next effect sets a virtual size to rely on OpenGL's
+				// bilinear sampling, we'll really need to break the phase here.
+				start_new_phase = true;
+			} else if (deps[i]->effect->changes_output_size() && !node->one_to_one_sampling) {
+				// If the next effect changes size and we don't have one-to-one sampling,
+				// we also need to break here.
 				start_new_phase = true;
 			}
 
@@ -367,6 +548,10 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 				phase->inputs.push_back(construct_phase(deps[i], completed_effects));
 			} else {
 				effects_todo_this_phase.push(deps[i]);
+
+				// Propagate the one-to-one status down through the dependency.
+				deps[i]->one_to_one_sampling = node->one_to_one_sampling &&
+					deps[i]->effect->one_to_one_sampling();
 			}
 		}
 	}
@@ -379,6 +564,9 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 	sort(phase->inputs.begin(), phase->inputs.end());
 	phase->inputs.erase(unique(phase->inputs.begin(), phase->inputs.end()), phase->inputs.end());
 
+	// Allocate samplers for each input.
+	phase->input_samplers.resize(phase->inputs.size());
+
 	// We added the effects from the output and back, but we need to output
 	// them in topological sort order in the shader.
 	phase->effects = topological_sort(phase->effects);
@@ -392,13 +580,30 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 	for (unsigned i = 0; i < phase->effects.size(); ++i) {
 		Node *node = phase->effects[i];
 		if (node->effect->num_inputs() == 0) {
-			CHECK(node->effect->set_int("needs_mipmaps", phase->input_needs_mipmaps));
+			Input *input = static_cast<Input *>(node->effect);
+			assert(!phase->input_needs_mipmaps || input->can_supply_mipmaps());
+			CHECK(input->set_int("needs_mipmaps", phase->input_needs_mipmaps));
 		}
 	}
 
+	// Tell each node which phase it ended up in, so that the unit test
+	// can check that the phases were split in the right place.
+	// Note that this ignores that effects may be part of multiple phases;
+	// if the unit tests need to test such cases, we'll reconsider.
+	for (unsigned i = 0; i < phase->effects.size(); ++i) {
+		phase->effects[i]->containing_phase = phase;
+	}
+
 	// Actually make the shader for this phase.
 	compile_glsl_program(phase);
 
+	// Initialize timer objects.
+	if (movit_timer_queries_supported) {
+		glGenQueries(1, &phase->timer_query_object);
+		phase->time_elapsed_ns = 0;
+		phase->num_measured_iterations = 0;
+	}
+
 	assert(completed_effects->count(output) == 0);
 	completed_effects->insert(make_pair(output, phase));
 	phases.push_back(phase);
@@ -593,7 +798,8 @@ void EffectChain::inform_input_sizes(Phase *phase)
 	// Now propagate from the inputs towards the end, and inform as we go.
 	// The rules are simple:
 	//
-	//   1. Don't touch effects that already have given sizes (ie., inputs).
+	//   1. Don't touch effects that already have given sizes (ie., inputs
+	//      or effects that change the output size).
 	//   2. If all of your inputs have the same size, that will be your output size.
 	//   3. Otherwise, your output size is 0x0.
 	for (unsigned i = 0; i < phase->effects.size(); ++i) {
@@ -615,8 +821,19 @@ void EffectChain::inform_input_sizes(Phase *phase)
 				this_output_height = 0;
 			}
 		}
-		node->output_width = this_output_width;
-		node->output_height = this_output_height;
+		if (node->effect->changes_output_size()) {
+			// We cannot call get_output_size() before we've done inform_input_size()
+			// on all inputs.
+			unsigned real_width, real_height;
+			node->effect->get_output_size(&real_width, &real_height,
+			                              &node->output_width, &node->output_height);
+			assert(node->effect->sets_virtual_output_size() ||
+			       (real_width == node->output_width &&
+			        real_height == node->output_height));
+		} else {
+			node->output_width = this_output_width;
+			node->output_height = this_output_height;
+		}
 	}
 }
 
@@ -630,6 +847,9 @@ void EffectChain::find_output_size(Phase *phase)
 	if (output_node->effect->changes_output_size()) {
 		output_node->effect->get_output_size(&phase->output_width, &phase->output_height,
 		                                     &phase->virtual_output_width, &phase->virtual_output_height);
+		assert(output_node->effect->sets_virtual_output_size() ||
+		       (phase->output_width == phase->virtual_output_width &&
+			phase->output_height == phase->virtual_output_height));
 		return;
 	}
 
@@ -1291,6 +1511,22 @@ void EffectChain::fix_output_gamma()
 		connect_nodes(output, conversion);
 	}
 }
+
+// If the user has requested Y'CbCr output, we need to do this conversion
+// _after_ GammaCompressionEffect etc., but before dither (see below).
+// This is because Y'CbCr, with the exception of a special optional mode
+// in Rec. 2020 (which we currently don't support), is defined to work on
+// gamma-encoded data.
+void EffectChain::add_ycbcr_conversion_if_needed()
+{
+	assert(output_color_type == OUTPUT_COLOR_RGB || output_color_type == OUTPUT_COLOR_YCBCR);
+	if (output_color_type != OUTPUT_COLOR_YCBCR) {
+		return;
+	}
+	Node *output = find_output_node();
+	Node *ycbcr = add_node(new YCbCrConversionEffect(output_ycbcr_format));
+	connect_nodes(output, ycbcr);
+}
 	
 // If the user has requested dither, add a DitherEffect right at the end
 // (after GammaCompressionEffect etc.). This needs to be done after everything else,
@@ -1373,11 +1609,13 @@ void EffectChain::finalize()
 	fix_internal_gamma_by_asking_inputs(15);
 	fix_internal_gamma_by_inserting_nodes(16);
 
-	output_dot("step17-before-dither.dot");
+	output_dot("step17-before-ycbcr.dot");
+	add_ycbcr_conversion_if_needed();
 
+	output_dot("step18-before-dither.dot");
 	add_dither_if_needed();
 
-	output_dot("step18-final.dot");
+	output_dot("step19-final.dot");
 	
 	// Construct all needed GLSL programs, starting at the output.
 	// We need to keep track of which effects have already been computed,
@@ -1386,7 +1624,7 @@ void EffectChain::finalize()
 	map<Node *, Phase *> completed_effects;
 	construct_phase(find_output_node(), &completed_effects);
 
-	output_dot("step19-split-to-phases.dot");
+	output_dot("step20-split-to-phases.dot");
 
 	assert(phases[0]->inputs.empty());
 	
@@ -1417,6 +1655,23 @@ void EffectChain::render_to_fbo(GLuint dest_fbo, unsigned width, unsigned height
 	glDepthMask(GL_FALSE);
 	check_error();
 
+	// Generate a VAO. All the phases should have exactly the same vertex attributes,
+	// so it's safe to reuse this.
+	float vertices[] = {
+		0.0f, 2.0f,
+		0.0f, 0.0f,
+		2.0f, 0.0f
+	};
+
+	GLuint vao;
+	glGenVertexArrays(1, &vao);
+	check_error();
+	glBindVertexArray(vao);
+	check_error();
+
+	GLuint position_vbo = fill_vertex_attribute(phases[0]->glsl_program_num, "position", 2, GL_FLOAT, sizeof(vertices), vertices);
+	GLuint texcoord_vbo = fill_vertex_attribute(phases[0]->glsl_program_num, "texcoord", 2, GL_FLOAT, sizeof(vertices), vertices);  // Same as vertices.
+
 	set<Phase *> generated_mipmaps;
 
 	// We choose the simplest option of having one texture per output,
@@ -1426,6 +1681,9 @@ void EffectChain::render_to_fbo(GLuint dest_fbo, unsigned width, unsigned height
 	for (unsigned phase_num = 0; phase_num < phases.size(); ++phase_num) {
 		Phase *phase = phases[phase_num];
 
+		if (do_phase_timing) {
+			glBeginQuery(GL_TIME_ELAPSED, phase->timer_query_object);
+		}
 		if (phase_num == phases.size() - 1) {
 			// Last phase goes to the output the user specified.
 			glBindFramebuffer(GL_FRAMEBUFFER, dest_fbo);
@@ -1439,6 +1697,9 @@ void EffectChain::render_to_fbo(GLuint dest_fbo, unsigned width, unsigned height
 			}
 		}
 		execute_phase(phase, phase_num == phases.size() - 1, &output_textures, &generated_mipmaps);
+		if (do_phase_timing) {
+			glEndQuery(GL_TIME_ELAPSED);
+		}
 	}
 
 	for (map<Phase *, GLuint>::const_iterator texture_it = output_textures.begin();
@@ -1451,6 +1712,63 @@ void EffectChain::render_to_fbo(GLuint dest_fbo, unsigned width, unsigned height
 	check_error();
 	glUseProgram(0);
 	check_error();
+
+	cleanup_vertex_attribute(phases[0]->glsl_program_num, "position", position_vbo);
+	cleanup_vertex_attribute(phases[0]->glsl_program_num, "texcoord", texcoord_vbo);
+
+	glDeleteVertexArrays(1, &vao);
+	check_error();
+
+	if (do_phase_timing) {
+		// Get back the timer queries.
+		for (unsigned phase_num = 0; phase_num < phases.size(); ++phase_num) {
+			Phase *phase = phases[phase_num];
+			GLint available = 0;
+			while (!available) {
+				glGetQueryObjectiv(phase->timer_query_object, GL_QUERY_RESULT_AVAILABLE, &available);
+			}
+			GLuint64 time_elapsed;
+			glGetQueryObjectui64v(phase->timer_query_object, GL_QUERY_RESULT, &time_elapsed);
+			phase->time_elapsed_ns += time_elapsed;
+			++phase->num_measured_iterations;
+		}
+	}
+}
+
+void EffectChain::enable_phase_timing(bool enable)
+{
+	if (enable) {
+		assert(movit_timer_queries_supported);
+	}
+	this->do_phase_timing = enable;
+}
+
+void EffectChain::reset_phase_timing()
+{
+	for (unsigned phase_num = 0; phase_num < phases.size(); ++phase_num) {
+		Phase *phase = phases[phase_num];
+		phase->time_elapsed_ns = 0;
+		phase->num_measured_iterations = 0;
+	}
+}
+
+void EffectChain::print_phase_timing()
+{
+	double total_time_ms = 0.0;
+	for (unsigned phase_num = 0; phase_num < phases.size(); ++phase_num) {
+		Phase *phase = phases[phase_num];
+		double avg_time_ms = phase->time_elapsed_ns * 1e-6 / phase->num_measured_iterations;
+		printf("Phase %d: %5.1f ms  [", phase_num, avg_time_ms);
+		for (unsigned effect_num = 0; effect_num < phase->effects.size(); ++effect_num) {
+			if (effect_num != 0) {
+				printf(", ");
+			}
+			printf("%s", phase->effects[effect_num]->effect->effect_type_id().c_str());
+		}
+		printf("]\n");
+		total_time_ms += avg_time_ms;
+	}
+	printf("Total:   %5.1f ms\n", total_time_ms);
 }
 
 void EffectChain::execute_phase(Phase *phase, bool last_phase, map<Phase *, GLuint> *output_textures, set<Phase *> *generated_mipmaps)
@@ -1483,7 +1801,8 @@ void EffectChain::execute_phase(Phase *phase, bool last_phase, map<Phase *, GLui
 			check_error();
 			generated_mipmaps->insert(input);
 		}
-		setup_rtt_sampler(glsl_program_num, sampler, phase->effect_ids[input->output_node], phase->input_needs_mipmaps);
+		setup_rtt_sampler(sampler, phase->input_needs_mipmaps);
+		phase->input_samplers[sampler] = sampler;  // Bind the sampler to the right uniform.
 	}
 
 	// And now the output. (Already set up for us if it is the last phase.)
@@ -1509,29 +1828,13 @@ void EffectChain::execute_phase(Phase *phase, bool last_phase, map<Phase *, GLui
 		}
 	}
 
-	// Now draw!
-	float vertices[] = {
-		0.0f, 1.0f,
-		0.0f, 0.0f,
-		1.0f, 1.0f,
-		1.0f, 0.0f
-	};
+	// Uniforms need to come after set_gl_state(), since they can be updated
+	// from there.
+	setup_uniforms(phase);
 
-	GLuint vao;
-	glGenVertexArrays(1, &vao);
-	check_error();
-	glBindVertexArray(vao);
+	glDrawArrays(GL_TRIANGLES, 0, 3);
 	check_error();
 
-	GLuint position_vbo = fill_vertex_attribute(glsl_program_num, "position", 2, GL_FLOAT, sizeof(vertices), vertices);
-	GLuint texcoord_vbo = fill_vertex_attribute(glsl_program_num, "texcoord", 2, GL_FLOAT, sizeof(vertices), vertices);  // Same as vertices.
-
-	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
-	check_error();
-
-	cleanup_vertex_attribute(glsl_program_num, "position", position_vbo);
-	cleanup_vertex_attribute(glsl_program_num, "texcoord", texcoord_vbo);
-	
 	glUseProgram(0);
 	check_error();
 
@@ -1543,12 +1846,71 @@ void EffectChain::execute_phase(Phase *phase, bool last_phase, map<Phase *, GLui
 	if (!last_phase) {
 		resource_pool->release_fbo(fbo);
 	}
-
-	glDeleteVertexArrays(1, &vao);
-	check_error();
 }
 
-void EffectChain::setup_rtt_sampler(GLuint glsl_program_num, int sampler_num, const string &effect_id, bool use_mipmaps)
+void EffectChain::setup_uniforms(Phase *phase)
+{
+	// TODO: Use UBO blocks.
+	for (size_t i = 0; i < phase->uniforms_sampler2d.size(); ++i) {
+		const Uniform<int> &uniform = phase->uniforms_sampler2d[i];
+		if (uniform.location != -1) {
+			glUniform1iv(uniform.location, uniform.num_values, uniform.value);
+		}
+	}
+	for (size_t i = 0; i < phase->uniforms_bool.size(); ++i) {
+		const Uniform<bool> &uniform = phase->uniforms_bool[i];
+		assert(uniform.num_values == 1);
+		if (uniform.location != -1) {
+			glUniform1i(uniform.location, *uniform.value);
+		}
+	}
+	for (size_t i = 0; i < phase->uniforms_int.size(); ++i) {
+		const Uniform<int> &uniform = phase->uniforms_int[i];
+		if (uniform.location != -1) {
+			glUniform1iv(uniform.location, uniform.num_values, uniform.value);
+		}
+	}
+	for (size_t i = 0; i < phase->uniforms_float.size(); ++i) {
+		const Uniform<float> &uniform = phase->uniforms_float[i];
+		if (uniform.location != -1) {
+			glUniform1fv(uniform.location, uniform.num_values, uniform.value);
+		}
+	}
+	for (size_t i = 0; i < phase->uniforms_vec2.size(); ++i) {
+		const Uniform<float> &uniform = phase->uniforms_vec2[i];
+		if (uniform.location != -1) {
+			glUniform2fv(uniform.location, uniform.num_values, uniform.value);
+		}
+	}
+	for (size_t i = 0; i < phase->uniforms_vec3.size(); ++i) {
+		const Uniform<float> &uniform = phase->uniforms_vec3[i];
+		if (uniform.location != -1) {
+			glUniform3fv(uniform.location, uniform.num_values, uniform.value);
+		}
+	}
+	for (size_t i = 0; i < phase->uniforms_vec4.size(); ++i) {
+		const Uniform<float> &uniform = phase->uniforms_vec4[i];
+		if (uniform.location != -1) {
+			glUniform4fv(uniform.location, uniform.num_values, uniform.value);
+		}
+	}
+	for (size_t i = 0; i < phase->uniforms_mat3.size(); ++i) {
+		const Uniform<Matrix3d> &uniform = phase->uniforms_mat3[i];
+		assert(uniform.num_values == 1);
+		if (uniform.location != -1) {
+			// Convert to float (GLSL has no double matrices).
+		        float matrixf[9];
+			for (unsigned y = 0; y < 3; ++y) {
+				for (unsigned x = 0; x < 3; ++x) {
+					matrixf[y + x * 3] = (*uniform.value)(y, x);
+				}
+			}
+			glUniformMatrix3fv(uniform.location, 1, GL_FALSE, matrixf);
+		}
+	}
+}
+
+void EffectChain::setup_rtt_sampler(int sampler_num, bool use_mipmaps)
 {
 	glActiveTexture(GL_TEXTURE0 + sampler_num);
 	check_error();
@@ -1563,10 +1925,6 @@ void EffectChain::setup_rtt_sampler(GLuint glsl_program_num, int sampler_num, co
 	check_error();
 	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
 	check_error();
-
-	string texture_name = string("tex_") + effect_id;
-	glUniform1i(glGetUniformLocation(glsl_program_num, texture_name.c_str()), sampler_num);
-	check_error();
 }
 
 }  // namespace movit
