@@ -1,4 +1,6 @@
 // Three-lobed Lanczos, the most common choice.
+// Note that if you change this, the accuracy for LANCZOS_TABLE_SIZE
+// needs to be recomputed.
 #define LANCZOS_RADIUS 3.0
 
 #include <epoxy/gl.h>
@@ -40,15 +42,69 @@ float sinc(float x)
 	}
 }
 
-float lanczos_weight(float x, float a)
+float lanczos_weight(float x)
 {
-	if (fabs(x) > a) {
+	if (fabs(x) > LANCZOS_RADIUS) {
 		return 0.0f;
 	} else {
-		return sinc(M_PI * x) * sinc(M_PI * x / a);
+		return sinc(M_PI * x) * sinc((M_PI / LANCZOS_RADIUS) * x);
 	}
 }
 
+// The weight function can be expensive to compute over and over again
+// (which will happen during e.g. a zoom), but it is also easy to interpolate
+// linearly. We compute the right half of the function (in the range of
+// 0..LANCZOS_RADIUS), with two guard elements for easier interpolation, and
+// linearly interpolate to get our function.
+//
+// We want to scale the table so that the maximum error is always smaller
+// than 1e-6. As per http://www-solar.mcs.st-andrews.ac.uk/~clare/Lectures/num-analysis/Numan_chap3.pdf,
+// the error for interpolating a function linearly between points [a,b] is
+//
+//   e = 1/2 (x-a)(x-b) f''(u_x)
+//
+// for some point u_x in [a,b] (where f(x) is our Lanczos function; we're
+// assuming LANCZOS_RADIUS=3 from here on). Obviously this is bounded by
+// f''(x) over the entire range. Numeric optimization shows the maximum of
+// |f''(x)| to be in x=1.09369819474562880, with the value 2.40067758733152381.
+// So if the steps between consecutive values are called d, we get
+//
+//   |e| <= 1/2 (d/2)^2 2.4007
+//   |e| <= 0.1367 d^2
+//
+// Solve for e = 1e-6 yields a step size of 0.0027, which to cover the range
+// 0..3 needs 1109 steps. We round up to the next power of two, just to be sure.
+#define LANCZOS_TABLE_SIZE 2048
+bool lanczos_table_init_done = false;
+float lanczos_table[LANCZOS_TABLE_SIZE + 2];
+
+void init_lanczos_table()
+{
+	for (unsigned i = 0; i < LANCZOS_TABLE_SIZE + 2; ++i) {
+		lanczos_table[i] = lanczos_weight(float(i) * (LANCZOS_RADIUS / LANCZOS_TABLE_SIZE));
+	}
+	lanczos_table_init_done = true;
+}
+
+float lanczos_weight_cached(float x)
+{
+	if (!lanczos_table_init_done) {
+		// Could in theory race between two threads if we are unlucky,
+		// but that is harmless, since they'll write the same data.
+		init_lanczos_table();
+	}
+	x = fabs(x);
+	if (x > LANCZOS_RADIUS) {
+		return 0.0f;
+	}
+	float table_pos = x * (LANCZOS_TABLE_SIZE / LANCZOS_RADIUS);
+	int table_pos_int = int(table_pos);  // Truncate towards zero.
+	float table_pos_frac = table_pos - table_pos_int;
+	assert(table_pos < LANCZOS_TABLE_SIZE + 2);
+	return lanczos_table[table_pos_int] +
+		table_pos_frac * (lanczos_table[table_pos_int + 1] - lanczos_table[table_pos_int]);
+}
+
 // Euclid's algorithm, from Wikipedia.
 unsigned gcd(unsigned a, unsigned b)
 {
@@ -61,7 +117,7 @@ unsigned gcd(unsigned a, unsigned b)
 }
 
 template<class DestFloat>
-unsigned combine_samples(const Tap<float> *src, Tap<DestFloat> *dst, unsigned src_size, unsigned num_src_samples, unsigned max_samples_saved)
+unsigned combine_samples(const Tap<float> *src, Tap<DestFloat> *dst, float num_subtexels, float inv_num_subtexels, unsigned num_src_samples, unsigned max_samples_saved)
 {
 	// Cut off near-zero values at both sides.
 	unsigned num_samples_saved = 0;
@@ -107,9 +163,9 @@ unsigned combine_samples(const Tap<float> *src, Tap<DestFloat> *dst, unsigned sr
 		float pos2 = src[i + 1].pos;
 		assert(pos2 > pos1);
 
-		fp16_int_t pos, total_weight;
+		DestFloat pos, total_weight;
 		float sum_sq_error;
-		combine_two_samples(w1, w2, pos1, pos2, src_size, &pos, &total_weight, &sum_sq_error);
+		combine_two_samples(w1, w2, pos1, pos2, num_subtexels, inv_num_subtexels, &pos, &total_weight, &sum_sq_error);
 
 		// If the interpolation error is larger than that of about sqrt(2) of
 		// a level at 8-bit precision, don't combine. (You'd think 1.0 was enough,
@@ -142,8 +198,9 @@ void normalize_sum(Tap<T>* vals, unsigned num)
 		for (unsigned i = 0; i < num; ++i) {
 			sum += to_fp64(vals[i].weight);
 		}
+		double inv_sum = 1.0 / sum;
 		for (unsigned i = 0; i < num; ++i) {
-			vals[i].weight = from_fp64<T>(to_fp64(vals[i].weight) / sum);
+			vals[i].weight = from_fp64<T>(to_fp64(vals[i].weight) * inv_sum);
 		}
 	}
 }
@@ -159,23 +216,28 @@ void normalize_sum(Tap<T>* vals, unsigned num)
 template<class DestFloat>
 unsigned combine_many_samples(const Tap<float> *weights, unsigned src_size, unsigned src_samples, unsigned dst_samples, Tap<DestFloat> **bilinear_weights)
 {
-	int src_bilinear_samples = 0;
-	for (unsigned y = 0; y < dst_samples; ++y) {
-		unsigned num_samples_saved = combine_samples<DestFloat>(weights + y * src_samples, NULL, src_size, src_samples, UINT_MAX);
-		src_bilinear_samples = max<int>(src_bilinear_samples, src_samples - num_samples_saved);
+	float num_subtexels = src_size / movit_texel_subpixel_precision;
+	float inv_num_subtexels = movit_texel_subpixel_precision / src_size;
+
+	unsigned max_samples_saved = UINT_MAX;
+	for (unsigned y = 0; y < dst_samples && max_samples_saved > 0; ++y) {
+		unsigned num_samples_saved = combine_samples<DestFloat>(weights + y * src_samples, NULL, num_subtexels, inv_num_subtexels, src_samples, max_samples_saved);
+		max_samples_saved = min(max_samples_saved, num_samples_saved);
 	}
 
 	// Now that we know the right width, actually combine the samples.
+	unsigned src_bilinear_samples = src_samples - max_samples_saved;
 	*bilinear_weights = new Tap<DestFloat>[dst_samples * src_bilinear_samples];
 	for (unsigned y = 0; y < dst_samples; ++y) {
 		Tap<DestFloat> *bilinear_weights_ptr = *bilinear_weights + y * src_bilinear_samples;
 		unsigned num_samples_saved = combine_samples(
 			weights + y * src_samples,
 			bilinear_weights_ptr,
-			src_size,
+			num_subtexels,
+			inv_num_subtexels,
 			src_samples,
-			src_samples - src_bilinear_samples);
-		assert(int(src_samples) - int(num_samples_saved) == src_bilinear_samples);
+			max_samples_saved);
+		assert(num_samples_saved == max_samples_saved);
 		normalize_sum(bilinear_weights_ptr, src_bilinear_samples);
 	}
 	return src_bilinear_samples;
@@ -380,7 +442,8 @@ SingleResamplePassEffect::SingleResamplePassEffect(ResampleEffect *parent)
 	  last_output_width(-1),
 	  last_output_height(-1),
 	  last_offset(0.0 / 0.0),  // NaN.
-	  last_zoom(0.0 / 0.0)  // NaN.
+	  last_zoom(0.0 / 0.0),  // NaN.
+	  last_texture_width(-1), last_texture_height(-1)
 {
 	register_int("direction", (int *)&direction);
 	register_int("input_width", &input_width);
@@ -389,6 +452,13 @@ SingleResamplePassEffect::SingleResamplePassEffect(ResampleEffect *parent)
 	register_int("output_height", &output_height);
 	register_float("offset", &offset);
 	register_float("zoom", &zoom);
+	register_uniform_sampler2d("sample_tex", &uniform_sample_tex);
+	register_uniform_int("num_samples", &uniform_num_samples);  // FIXME: What about GLSL pre-1.30?
+	register_uniform_float("num_loops", &uniform_num_loops);
+	register_uniform_float("slice_height", &uniform_slice_height);
+	register_uniform_float("sample_x_scale", &uniform_sample_x_scale);
+	register_uniform_float("sample_x_offset", &uniform_sample_x_offset);
+	register_uniform_float("whole_pixel_offset", &uniform_whole_pixel_offset);
 
 	glGenTextures(1, &texnum);
 }
@@ -517,7 +587,7 @@ void SingleResamplePassEffect::update_texture(GLuint glsl_program_num, const str
 		// Now sample <int_radius> pixels on each side around that point.
 		for (int i = 0; i < src_samples; ++i) {
 			int src_y = base_src_y + i - int_radius;
-			float weight = lanczos_weight(radius_scaling_factor * (src_y - center_src_y - subpixel_offset), LANCZOS_RADIUS);
+			float weight = lanczos_weight_cached(radius_scaling_factor * (src_y - center_src_y - subpixel_offset));
 			weights[y * src_samples + i].weight = weight * radius_scaling_factor;
 			weights[y * src_samples + i].pos = (src_y + 0.5) / float(src_size);
 		}
@@ -554,16 +624,39 @@ void SingleResamplePassEffect::update_texture(GLuint glsl_program_num, const str
 	check_error();
 	glBindTexture(GL_TEXTURE_2D, texnum);
 	check_error();
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
-	check_error();
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
-	check_error();
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
-	check_error();
+	if (last_texture_width == -1) {
+		// Need to set this state the first time.
+		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
+		check_error();
+		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
+		check_error();
+		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
+		check_error();
+	}
+
+	GLenum type, internal_format;
+	void *pixels;
 	if (fallback_to_fp32) {
-		glTexImage2D(GL_TEXTURE_2D, 0, GL_RG32F, src_bilinear_samples, dst_samples, 0, GL_RG, GL_FLOAT, bilinear_weights_fp32);
+		type = GL_FLOAT;
+		internal_format = GL_RG32F;
+		pixels = bilinear_weights_fp32;
 	} else {
-		glTexImage2D(GL_TEXTURE_2D, 0, GL_RG16F, src_bilinear_samples, dst_samples, 0, GL_RG, GL_HALF_FLOAT, bilinear_weights_fp16);
+		type = GL_HALF_FLOAT;
+		internal_format = GL_RG16F;
+		pixels = bilinear_weights_fp16;
+	}
+
+	if (int(src_bilinear_samples) == last_texture_width &&
+	    int(dst_samples) == last_texture_height &&
+	    internal_format == last_texture_internal_format) {
+		// Texture dimensions and type are unchanged; it is more efficient
+		// to just update it rather than making an entirely new texture.
+		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, src_bilinear_samples, dst_samples, GL_RG, type, pixels);
+	} else {
+		glTexImage2D(GL_TEXTURE_2D, 0, internal_format, src_bilinear_samples, dst_samples, 0, GL_RG, type, pixels);
+		last_texture_width = src_bilinear_samples;
+		last_texture_height = dst_samples;
+		last_texture_internal_format = internal_format;
 	}
 	check_error();
 
@@ -601,23 +694,21 @@ void SingleResamplePassEffect::set_gl_state(GLuint glsl_program_num, const strin
 	glBindTexture(GL_TEXTURE_2D, texnum);
 	check_error();
 
-	set_uniform_int(glsl_program_num, prefix, "sample_tex", *sampler_num);
+	uniform_sample_tex = *sampler_num;
 	++*sampler_num;
-	set_uniform_int(glsl_program_num, prefix, "num_samples", src_bilinear_samples);
-	set_uniform_float(glsl_program_num, prefix, "num_loops", num_loops);
-	set_uniform_float(glsl_program_num, prefix, "slice_height", slice_height);
+	uniform_num_samples = src_bilinear_samples;
+	uniform_num_loops = num_loops;
+	uniform_slice_height = slice_height;
 
 	// Instructions for how to convert integer sample numbers to positions in the weight texture.
-	set_uniform_float(glsl_program_num, prefix, "sample_x_scale", 1.0f / src_bilinear_samples);
-	set_uniform_float(glsl_program_num, prefix, "sample_x_offset", 0.5f / src_bilinear_samples);
+	uniform_sample_x_scale = 1.0f / src_bilinear_samples;
+	uniform_sample_x_offset = 0.5f / src_bilinear_samples;
 
-	float whole_pixel_offset;
 	if (direction == SingleResamplePassEffect::VERTICAL) {
-		whole_pixel_offset = lrintf(offset) / float(input_height);
+		uniform_whole_pixel_offset = lrintf(offset) / float(input_height);
 	} else {
-		whole_pixel_offset = lrintf(offset) / float(input_width);
+		uniform_whole_pixel_offset = lrintf(offset) / float(input_width);
 	}
-	set_uniform_float(glsl_program_num, prefix, "whole_pixel_offset", whole_pixel_offset);
 
 	// We specifically do not want mipmaps on the input texture;
 	// they break minification.
