@@ -156,6 +156,7 @@ GLuint compile_shader(const string &shader_src, GLenum type)
 	GLint status;
 	glGetShaderiv(obj, GL_COMPILE_STATUS, &status);
 	if (status == GL_FALSE) {
+		fprintf(stderr, "Failed to compile shader: %s\n", shader_src.c_str());
 		exit(1);
 	}
 
@@ -176,6 +177,7 @@ string output_glsl_mat3(const string &name, const Eigen::Matrix3d &m)
 	stringstream ss;
 	ss.imbue(locale("C"));
 	ss.precision(8);
+	ss << scientific;
 	ss << "const mat3 " << name << " = mat3(\n";
 	ss << "    " << m(0,0) << ", " << m(1,0) << ", " << m(2,0) << ",\n";
 	ss << "    " << m(0,1) << ", " << m(1,1) << ", " << m(2,1) << ",\n";
@@ -183,12 +185,24 @@ string output_glsl_mat3(const string &name, const Eigen::Matrix3d &m)
 	return ss.str();
 }
 
+string output_glsl_float(const string &name, float x)
+{
+	// Use stringstream to be independent of the current locale in a thread-safe manner.
+	stringstream ss;
+	ss.imbue(locale("C"));
+	ss.precision(8);
+	ss << scientific;
+	ss << "const float " << name << " = " << x << ";\n";
+	return ss.str();
+}
+
 string output_glsl_vec2(const string &name, float x, float y)
 {
 	// Use stringstream to be independent of the current locale in a thread-safe manner.
 	stringstream ss;
 	ss.imbue(locale("C"));
 	ss.precision(8);
+	ss << scientific;
 	ss << "const vec2 " << name << " = vec2(" << x << ", " << y << ");\n";
 	return ss.str();
 }
@@ -199,12 +213,13 @@ string output_glsl_vec3(const string &name, float x, float y, float z)
 	stringstream ss;
 	ss.imbue(locale("C"));
 	ss.precision(8);
+	ss << scientific;
 	ss << "const vec3 " << name << " = vec3(" << x << ", " << y << ", " << z << ");\n";
 	return ss.str();
 }
 
 template<class DestFloat>
-void combine_two_samples(float w1, float w2, float pos1, float pos2, unsigned size,
+void combine_two_samples(float w1, float w2, float pos1, float pos2, float num_subtexels, float inv_num_subtexels,
                          DestFloat *offset, DestFloat *total_weight, float *sum_sq_error)
 {
 	assert(movit_initialized);
@@ -223,9 +238,7 @@ void combine_two_samples(float w1, float w2, float pos1, float pos2, unsigned si
 	// Round to the minimum number of bits we have measured earlier.
 	// The card will do this for us anyway, but if we know what the real z
 	// is, we can pick a better total_weight below.
-	z *= size;  // Move to pixel coordinates,
-	z = lrintf(z / movit_texel_subpixel_precision) * movit_texel_subpixel_precision;  // Round.
-	z /= size;  // Move back to normalized coordinates.
+	z = lrintf(z * num_subtexels) * inv_num_subtexels;
 	
 	// Choose total weight w so that we minimize total squared error
 	// for the effective weights:
@@ -240,22 +253,22 @@ void combine_two_samples(float w1, float w2, float pos1, float pos2, unsigned si
 	//   w = (a(1-z) + bz) / ((1-z)² + z²)
 	//
 	// If z had infinite precision, this would simply reduce to w = w1 + w2.
-	*total_weight = (w1 * (1 - z) + w2 * z) / (z * z + (1 - z) * (1 - z));
+	*total_weight = from_fp64<DestFloat>((w1 + z * (w2 - w1)) / (z * z + (1 - z) * (1 - z)));
 
 	if (sum_sq_error != NULL) {
-		float err1 = *total_weight * (1 - z) - w1;
-		float err2 = *total_weight * z - w2;
+		float err1 = to_fp64(*total_weight) * (1 - z) - w1;
+		float err2 = to_fp64(*total_weight) * z - w2;
 		*sum_sq_error = err1 * err1 + err2 * err2;
 	}
 }
 
 // Explicit instantiations.
 template
-void combine_two_samples<float>(float w1, float w2, float pos1, float pos2, unsigned size,
+void combine_two_samples<float>(float w1, float w2, float pos1, float pos2, float num_subtexels, float inv_num_subtexels,
                                 float *offset, float *total_weight, float *sum_sq_error);
 
 template
-void combine_two_samples<fp16_int_t>(float w1, float w2, float pos1, float pos2, unsigned size,
+void combine_two_samples<fp16_int_t>(float w1, float w2, float pos1, float pos2, float num_subtexels, float inv_num_subtexels,
                                      fp16_int_t *offset, fp16_int_t *total_weight, float *sum_sq_error);
 
 GLuint fill_vertex_attribute(GLuint glsl_program_num, const string &attribute_name, GLint size, GLenum type, GLsizeiptr data_size, const GLvoid *data)
